/**
  Framebuffer Object (FBO) Demo Application
  
  Orion Sky Lawlor, olawlor@acm.org, 2008-04-10 (Public Domain)
*/
#include <GL/glew.h> /*<- for gl...ARB extentions.  Must call glewInit after glutCreateWindow! */
#include <iostream>
#include <GL/glut.h>
#include <math.h>
#include "osl/mat4.h"
using osl::mat4;
#include "osl/vec4.h" /*<- handy little vector class */
/* Convert a struct with .x, .y, .z fields into three arguments */
#define VEC3_TO_XYZ(v) (v).x,(v).y,(v).z

#include "GL/glui.h" /* user interface */

#include "ogl/image.h" /* (originally from) Nigel Stewart's GLT image library */
#include "ogl/image.cpp"

#include "ogl/glsl.h" /* Orion's tiny GLSL library */
#include "ogl/glsl.cpp"

int show_maps=1;
/* IFS parameters used below */
mat4 mat1(1.0), mat2(1.0), mat3(1.0); 
float scl1=1.0,scl2=1.0, scl3=1.0;

/*********************** App Code **********************/

/* Orthonormal coordinate frame */
class ortho_frame {
public:
	/* Unit vectors pointing along axes of our frame.
		X cross Y is Z. (right-handed coordinate system)
	*/
	vec3 x,y,z;
	ortho_frame() :x(1,0,0), y(0,1,0), z(0,0,1) {}
	
	/* Reorient this coordinate frame by this far 
	   along in the X and Y axes.  "dx" is the distance
	   along the z axis to push the x axis; "dy" 
	   the distance along the z axis to push the y axis.
	*/
	void nudge(double dx,double dy) {
		x+=dx*z;
		y+=dy*z;
		orthonormalize();
	}
	
	/* Reconstruct orthonormal frame from X and Y axes.
	  Y is primary, X is secondary, Z is tertiary.
	*/
	void orthonormalize(void) {
		y=normalize(y);
		z=normalize(cross(x,y));
		x=normalize(cross(y,z));
	}
};

/* Orientation of our camera. Camera looks down -Z axis; Y is up. */
ortho_frame camera_frame;

void oglExit(char *where,char *why) {
	printf("Fatal error> %s\n",why);
	exit(1);
}
void bad_framebuffer(GLenum status) {
        switch(status) {
        case GL_FRAMEBUFFER_COMPLETE_EXT:
                oglExit("oglFramebuffer::bind","framebuffer is actually OK (?!)");
                break;
        case GL_FRAMEBUFFER_UNSUPPORTED_EXT:
                oglExit("oglFramebuffer::bind","combination of formats is UNSUPPORTED by your card");
        case GL_FRAMEBUFFER_INCOMPLETE_ATTACHMENT_EXT:
                oglExit("oglFramebuffer::bind","ATTACHMENT_EXT");
        case GL_FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT_EXT:
                oglExit("oglFramebuffer::bind","MISSING_ATTACHMENT_EXT");
        case GL_FRAMEBUFFER_INCOMPLETE_DUPLICATE_ATTACHMENT_EXT:
                oglExit("oglFramebuffer::bind","DUPLICATE_ATTACHMENT_EXT");
        case GL_FRAMEBUFFER_INCOMPLETE_DIMENSIONS_EXT:
                oglExit("oglFramebuffer::bind","DIMENSIONS_EXT");
        case GL_FRAMEBUFFER_INCOMPLETE_FORMATS_EXT:
                oglExit("oglFramebuffer::bind","FORMATS_EXT");
        case GL_FRAMEBUFFER_INCOMPLETE_DRAW_BUFFER_EXT:
                oglExit("oglFramebuffer::bind","DRAW_BUFFER_EXT");
        case GL_FRAMEBUFFER_INCOMPLETE_READ_BUFFER_EXT:
                oglExit("oglFramebuffer::bind","READ_BUFFER_EXT");
        default:
                oglExit("oglFramebuffer::bind","GL_FRAMEBUFFER  logic error");
        };
}

/* Make the current texture "work right", with the usual filtering */
void tex_workright(void) {
	/* Texture state applies to the currently bound texture */
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_CLAMP_TO_EDGE);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_CLAMP_TO_EDGE);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAX_ANISOTROPY_EXT,16);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,
		GL_LINEAR_MIPMAP_LINEAR);
	//glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
	//glTexParameteri(GL_TEXTURE_2D,GL_GENERATE_MIPMAP_SGIS, GL_TRUE);
}

vec3 operator*(const mat4 &m,const vec3 &v) {
	return vec3(m*vec4(v,1));
}
void draw_texture(mat4 m,vec3 lo,vec3 hi)
{
	glBegin (GL_TRIANGLE_FAN);
        glTexCoord2f(0,0); glVertex3fv(m*vec3(lo.x,lo.y,lo.z)); 
        glTexCoord2f(1,0); glVertex3fv(m*vec3(hi.x,lo.y,lo.z)); 
        glTexCoord2f(1,1); glVertex3fv(m*vec3(hi.x,hi.y,lo.z)); 
        glTexCoord2f(0,1); glVertex3fv(m*vec3(lo.x,hi.y,lo.z)); 
        glEnd();
}

void my_display(void) {
	glFinish(); int last_msec=glutGet(GLUT_ELAPSED_TIME);
	
	glMatrixMode(GL_PROJECTION); /* clear out projection matrix */
	glLoadIdentity(); /* reset matrix at start of frame */
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	
	/* Framebuffer output texture */
	static GLuint frameTex[2]={0,0};
	int w=1024,h=1024; /* size of our textures */
	if (frameTex[0]==0) { 
		glGenTextures(2,frameTex); /* make a texture handle */
		for (int i=0;i<2;i++) {
			glBindTexture(GL_TEXTURE_2D,frameTex[i]);
			glTexImage2D(GL_TEXTURE_2D,0,
				GL_RGBA8,w,h, /* data format and size (pixels) */
				0,GL_LUMINANCE,GL_FLOAT,0);
			glGenerateMipmapEXT(GL_TEXTURE_2D); /*<- needed on many cards! */
			tex_workright();
		}
	}
	
	/* Framebuffer object */
	static GLuint fbo=0;
	if (fbo==0) { /* create up framebuffer object--set it up below */
		glGenFramebuffersEXT(1,&fbo);
	}
	
	/* Render into our framebuffer object */
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT,fbo);
	glViewport(0,0,w,h);
	glDisable(GL_DEPTH_TEST); /* or else attach to GL_DEPTH_ATTACHMENT_EXT! */
	glEnable(GL_BLEND);
	glBlendFunc(GL_ZERO,GL_SRC_COLOR); /* multiplicative blending--darken only */
	glScalef(1.0,1.0,0.01); /* scale down Z axis, to avoid z-clipping */
	
	/* Iteratively draw the IFS maps */
	for (int depth=0;depth<10;depth++)
	{
		// Make sure we're rendering into frameTex[1]
		glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT,
			GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_2D,
			frameTex[1], 0);
		/* Make sure the framebuffer's happy */
		GLenum status = glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT);
		if (status!=GL_FRAMEBUFFER_COMPLETE_EXT)
			bad_framebuffer(status);
		
		// Clear the screen
		glClearColor(1.0,1.0,1.0,0.0); /* transparent white */
		if (depth<=0) glClearColor(0.0,0.0,0.0,1.0); /* opaque black first time */
		glClear(GL_COLOR_BUFFER_BIT);
		
		// Draw three copies of texture frameTex[0]
		glBindTexture(GL_TEXTURE_2D,frameTex[0]);
		glEnable(GL_TEXTURE_2D);
		double g=0.9; /* greyscale darkening with each level */
		if (show_maps) glColor3f(g,g,0);
		draw_texture(scl1*mat1,vec3(-0.5, 0.0,0.0),vec3(0.5,1.0,0.0));
		draw_texture(scl2*mat2,vec3(-1.0,-1.0,0.0),vec3(0.0));
		draw_texture(scl3*mat3,vec3( 0.0,-1.0,0.0),vec3(1.0,0.0,0.0));
	
		// Build mipmaps of our freshly-rendered data
		glBindTexture(GL_TEXTURE_2D,frameTex[1]);
		glGenerateMipmapEXT(GL_TEXTURE_2D); 
		
		// Swap textures zero and one
		std::swap(frameTex[0],frameTex[1]);
	}
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT,0);
	glViewport(0,0,glutGet(GLUT_WINDOW_WIDTH),glutGet(GLUT_WINDOW_HEIGHT));
	glEnable(GL_DEPTH_TEST);
	glDisable(GL_BLEND);

/* Do normal fullscreen rendering */
	glMatrixMode(GL_PROJECTION); /* set up projection matrix */
	glLoadIdentity(); /* reset matrix at start of frame */
	gluPerspective(90.0, /* <- Y camera field-of-view, in degrees */
		glutGet(GLUT_WINDOW_WIDTH)/(float)glutGet(GLUT_WINDOW_HEIGHT), /* aspect ratio: width over height */
		0.01, /* Near clipping plane depth */
		100.0 /* Far clipping plane depth */
	);
	vec3 camera_location=0.7*camera_frame.z; /* back away from origin */
	gluLookAt(VEC3_TO_XYZ(vec3(0.5)+camera_location),
		VEC3_TO_XYZ(vec3(0.5)),
		VEC3_TO_XYZ(camera_frame.y));
	
	glMatrixMode(GL_MODELVIEW); /* set up model and view matrix */
	glLoadIdentity(); /* reset matrix at start of frame */
	
	glClearColor(0.3,0.5,0.7,0); /* set background color to blue */
	glClear(GL_COLOR_BUFFER_BIT + GL_DEPTH_BUFFER_BIT);
	glEnable(GL_DEPTH_TEST);
	
	/* Draw the current fixed-function texture */
	glBindTexture(GL_TEXTURE_2D,frameTex[0]);
	glEnable(GL_TEXTURE_2D); 
	glColor4f(1,1,1,1);
	draw_texture(mat4(1.0), vec3(0.0,0.0,0.5), vec3(1.0,1.0,0.5));

	glutSwapBuffers(); /* copy back buffer to front */
	
/* Statistics */
	glFinish();int cur_msec=glutGet(GLUT_ELAPSED_TIME); 
	int msec_elapsed=cur_msec-last_msec;
	printf("%d msec/frame; %.2f fps\n",
		msec_elapsed,1.0e3/msec_elapsed);
	last_msec=cur_msec;
}

int mouse_x, mouse_y;
void my_mouse(int state,int buttons,int x,int y) {
	mouse_x=x; mouse_y=y;  /* just stash motion */
}
void my_motion(int x,int y) { 
	float dx=x-mouse_x, dy=y-mouse_y; /* relative motion */
	
	float speed=1.0/100; /* radians rotation per pixel */
	camera_frame.nudge(speed*dx,-speed*dy); /* OpenGL Y is negative of mouse Y! */
	
	mouse_x=x; mouse_y=y; 
	glutPostRedisplay();
}

int main(int argc,char *argv[]) {
	glutInitWindowSize(800,600);
	glutInit(&argc,argv);
	int mode=GLUT_DOUBLE | GLUT_RGBA | GLUT_DEPTH;
        glutInitDisplayMode (mode);
	int win=glutCreateWindow("GLSL Framebuffer Object / IFS Demo");
	glewInit();
	
	glutDisplayFunc(my_display);
	glutMotionFunc(my_motion);
	glutMouseFunc(my_mouse);
	
	
	GLUI *glui = GLUI_Master.create_glui_subwindow(win, GLUI_SUBWINDOW_TOP);
	glui->set_main_gfx_window( win );
	
	new GLUI_Checkbox(glui,"Show Maps",&show_maps);	
	new GLUI_Column(glui);
	new GLUI_Rotation(glui,"Mat1",&mat1[0][0]);
	new GLUI_Spinner(glui,"Scl1",&scl1);
	new GLUI_Column(glui);
	new GLUI_Rotation(glui,"Mat2",&mat2[0][0]);
	new GLUI_Spinner(glui,"Scl2",&scl2);
	new GLUI_Column(glui);
	new GLUI_Rotation(glui,"Mat3",&mat3[0][0]);
	new GLUI_Spinner(glui,"Scl3",&scl3);
	
	glutMainLoop(); /* processes events */
	return 0;
}
